# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/0. Download models.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/0. Download models.ipynb 1
from fastcore.script import call_parse
import whisperx
import whisper

# %% ../nbs/0. Download models.ipynb 3
@call_parse
def main():
    whisper.load_model('base.en')
    whisper.load_model('small.en')
    whisperx.vad.load_vad_model('cpu')
    whisperx.asr.load_model('medium.en', "cpu", compute_type="float16", language='en')
